import(Module_String);
import(Module_Draw);
import(Module_Math);
import(Module_System);
import(Module_Table);
import(Module_Globals);

TextBox = {};
TextBox.__index = TextBox;

local gns = gnsi();

function TextBox:Create(_x, _y, _width, _height)
  local self = setmetatable({}, TextBox);

  self.X = _x;
  self.Y = _y;
  self.Width = _width or 240;
  self.Height = _height;
  self.Header = "";
  self.FormattedLines = {};

  return self;
end

function TextBox:SetHeaderText(str)
  self.Header = str;
end

function TextBox:FormatString(str)
  local curr_line = "";
  local curr_width = 0;
  PopSetFont(3);
  for word in str:gmatch("%S+") do
    if ((curr_width + string_width(word)) < self.Width) then
      curr_line = curr_line .. " " .. word;
      curr_width = curr_width + string_width(word);
    else
      table.insert(self.FormattedLines, curr_line);
      curr_line = word;
      curr_width = 0;
    end
  end

  -- Now figure the correct Y position for text.
  table.insert(self.FormattedLines, curr_line);
end

function TextBox:Render()
  --local str_test = string.format("ScreenWidth: %i, ScreenHeight: %i", gnsi.ScreenW, gnsi.ScreenH);
  --log(str_test);
  local gui_width = GFGetGuiWidth();
  PopSetFont(3);
  local y = math.floor(gns.ScreenH/2);
  local text_height = CharHeight2() * #self.FormattedLines;
  y = y - math.floor(text_height/2) - CharHeight2();
  self.Y = y;
  PopSetFont(9);
  LbDraw_Text((math.floor(gns.ScreenW / 2) - math.floor(string_width(self.Header) / 2)) + math.floor(gui_width / 2), self.Y-CharHeight2(), self.Header, 0);
  PopSetFont(3);
  for i,str in ipairs(self.FormattedLines) do
    LbDraw_Text((math.floor(gns.ScreenW / 2) - math.floor(string_width(str) / 2)) + math.floor(gui_width / 2), self.Y+(i*(CharHeight2())), str, 0);
  end
end

--Example on usage
--local briefing_msg = TextBox:Create(math.floor(gnsi.ScreenW/2), 100, 320, 240);
--briefing_msg:FormatString(read_me_text);
--briefing_msg:SetHeaderText("Luck of The Draw");

--function OnFrame()
  --briefing_msg:Render();
--end
