import(Module_System)
import(Module_Objects)
import(Module_Globals)
import(Module_Players)
import(Module_Map)
import(Module_Draw)
import(Module_Table)
import(Module_Game)
import(Module_Math)
import(Module_Level)
import(Module_DataTypes)
import(Module_PopScript)
import(Module_String)

log("Loading script...")

include("TextMsg.lua");

local gs = gsi()
local gns = gnsi()
local spell_info = spells_type_info()
local person_info = people_type_info()
local zones_buffer = {}

local player_info ={
  {"",0,0,887},
  {"",1,0,914},
  {"",2,0,941},
  {"",3,0,968},
  {"",4,0,1623},
  {"",5,0,1650},
  {"",6,0,1677},
  {"",7,0,1704}
}

player_scores = {0,0,0,0,0,0,0,0}
player_nums = {}
player_names = {}
spark_a_clrs = {2,-2,3,1,10,11,12,-1}
local game_state = 0
local init_turn = gs.Counts.ProcessThings
local message_delay = 48
local round_time = 720*5
local extra_pop_int = 2
local extra_mana_gen_int = 82
local mi = 1
local read_me_text = "Welcome to the sparkling tiles mini-game! Objective of this little event is to claim zones which are highlighted by sparkles, walk over them with shaman and they'll change their color to Your tribe. Each zone gives +1 point every second. The more zones You claim the more points You get.";
local ready_time = 12*20;
local tut_messages = {
  "Welcome to the capture zones mini-game!",
  "Objective of this little event is to claim zones which are highlighted by sparkles",
  "Walk over them with shaman and they'll change their color to Your tribe",
  "Each zone gives +1 point every second.",
  "The more zones You claim the more points You get.",
  "Get ready to claim zones!"
}

local welcomeBox = TextBox:Create(0, 0, 320, 240);
welcomeBox:FormatString(read_me_text);
welcomeBox:SetHeaderText("Sparkling Tiles");

person_info[7].ManaValue = 377
spell_info[17].WorldCoordRange = 2048
spell_info[4].Cost = 50000
spell_info[4].OneOffMaximum = 2
spell_info[3].Cost = 80000
spell_info[3].OneOffMaximum = 1

gns.GameParams.Flags2 = gns.GameParams.Flags2 | (1<<30)

local function toggle_player_spell(_pn,_spell,_arg)
  local ref = gs.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable
  if (_arg) then
    gs.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable = ref | (1 << _spell)
  elseif (not _arg) then
    gs.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable = ref ~ (1 << _spell)
  end
end

for i = 0,7 do
  if (gs.Players[i].DeadCount == 0) then
    local n = get_player_name(i,true)
    player_info[i+1][1] = n
    toggle_player_spell(i,17,true)
    SET_NO_REINC(i,TRUE)
    table.insert(player_names,n)
    table.insert(player_nums,i)
  end
end

local function toggle_player_bldg(_pn,_bldg,_arg)
  local ref = gs.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable
  if (_arg) then
    gs.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable = ref | (1 << _bldg)
  elseif (not _arg) then
    gs.ThisLevelInfo.PlayerThings[_pn].BuildingsAvailable = ref ~ (1 << _bldg)
  end
end

local function register_zone(c3d,r,_TABL)
  local mapXZ = MapPosXZ.new()
  mapXZ.Pos = world_coord3d_to_map_idx(c3d)
  mapXZ.XZ.X = mapXZ.XZ.X-r+1
  mapXZ.XZ.Z = mapXZ.XZ.Z-r+1
  local x = 0
  local z = 0
  local count = 0
  for i=0,r do
    x = mapXZ.XZ.X + i*2
    for i=0,r do
      z = mapXZ.XZ.Z + i*2
      local c3d_local = Coord3D.new()
      c3d_local = MAP_XZ_2_WORLD_XYZ(x,z)
      centre_coord3d_on_block(c3d_local)
      local zone_trail = createThing(7,61,8,c3d_local,false,false)
      zone_trail.u.Effect.Duration = -1
      zone_trail.Pos.D3.Ypos = zone_trail.Pos.D3.Ypos - 128
      table.insert(_TABL,zone_trail)
      local me = world_coord3d_to_map_ptr(c3d_local)
      --is spot walkable?
    end
  end
end

local coord_large = MAP_XZ_2_WORLD_XYZ(220, 226)
centre_coord3d_on_block(coord_large)
register_zone(coord_large,28,zones_buffer)

local respawn_spots = {
  MAP_XZ_2_WORLD_XYZ(222, 158),
  MAP_XZ_2_WORLD_XYZ(224, 40),
  MAP_XZ_2_WORLD_XYZ(30, 224),
  MAP_XZ_2_WORLD_XYZ(156, 232),
  MAP_XZ_2_WORLD_XYZ(14, 184),
  MAP_XZ_2_WORLD_XYZ(184, 16),
  MAP_XZ_2_WORLD_XYZ(172, 184),
  MAP_XZ_2_WORLD_XYZ(2, 14)
}

for i,c3d in ipairs(respawn_spots) do
  centre_coord3d_on_block(c3d)
end

local spells_to_give = {2,4,3}
local has_given_stuff = false

local function decide_winner()
  local ret = -1
  local score = 0
  for i,pnt in ipairs(player_info) do
    if (player_info[i][3] > score) then
      score = player_info[i][3]
      ret = player_info[i][2]
    end
  end
  return ret
end

local function add_score(pn,amt)
  for i in ipairs(player_info) do
    --Make sure playerNum matches with table's one.
    if (player_info[i][2] == pn) then
      player_info[i][3] = player_info[i][3] + amt
      break
    end
  end
end

local function update_leaderboard()
  --Sort by playerNum first then by score num.
  table.sort(player_info, function(a, b) if (a[3] ~= b[3]) then return a[3] < b[3] else return a[2] < b[2] end end)
end

function OnTurn()
  if (gs.Counts.ProcessThings > init_turn + 36) then
    if (game_state == 0) then
      if (ready_time > 0) then
        ready_time = ready_time - 1;
      else
        -- for i,pn in ipairs(player_nums) do
        --   createThing(7,49,8,player_spawn_coord[i],false,false)
        --   toggle_player_spell(pn,2,true)
        -- end
        game_state = 1
      end
      -- if (gs.Counts.ProcessThings > init_turn + 36 + (message_delay * mi)) then
      --   if (mi <= #tut_messages) then
      --     log_msg(8,tut_messages[mi])
      --     mi = mi+1
      --   else
      --     game_state = 1
      --   end
      -- end
    elseif (game_state == 1) then
      if (round_time > 0) then
        round_time = round_time - 1
        if (not has_given_stuff) then
          has_given_stuff = true
          for i,n in ipairs(spells_to_give) do
            for j=0,7 do
              toggle_player_spell(j,n,true)
            end
          end
        end

        if (gs.Counts.ProcessThings % 12 == 0) then
          for i,t in ipairs(zones_buffer) do
            if (t.Owner ~= 8) then
              gs.Players[t.Owner].ManaUnspent = gs.Players[t.Owner].ManaUnspent + extra_mana_gen_int
              --player_scores[t.Owner+1] = player_scores[t.Owner+1] + 1
              add_score(t.Owner,1)
            end
          end
        end

        if (gs.Counts.ProcessThings % 4 == 0) then
          update_leaderboard()
        end

        if (gs.Counts.ProcessThings % 3 == 0) then
          for i=0,7 do
            if (gs.Players[i].DeadCount == 0) then
              local s = getShaman(i)
              if (s ~= nil) then
                if (s.Flags2 & (1 << 10) == 0 and s.Flags & (1<<1) == 0) then
                  local m = world_coord2d_to_map_ptr(s.Pos.D2)
                  m.MapWhoList:processList(function(t)
                    if (t.Type == 7) then
                      if (t.Model == 61) then
                        t.DrawInfo.Alpha = spark_a_clrs[s.Owner+1]
                        t.Owner = s.Owner
                        return false
                      end
                    end
                    return true
                  end)
                end
              elseif (s == nil) then
                local spot = respawn_spots[i+1]
                createThing(1,7,i,spot,false,false)
              end
            end
          end
        end
      else
        game_state = 2
      end
    elseif (game_state == 2) then
      local winner = decide_winner()
      if (gns.PlayerNum == winner) then
        gns.Flags = gns.Flags | (1<<25)
      else
        gns.Flags = gns.Flags | (1<<26)
      end
      log_msg(8,"Time has ran out! Winner is: " .. player_names[winner+1] .. " - Congratulations!")
      game_state = 3621
    end
  end
end

function OnCreateThing(t)
  if (t.Type == 10) then
    if (t.Model == 12) then
      if (t.u.SoulConvert.CurrModel == 7) then
        t.SubState = 4
        t.u.SoulConvert.Count = 64
        t.DrawInfo.Flags = t.DrawInfo.Flags | (1<<4)
      end
    end
  end
end

--[[ took this piece of code from here coz i was lazy to write one myself
https://gist.github.com/jesseadams/791673
]]
local function SecondsToClock(seconds)
  local seconds = tonumber(seconds)

  if seconds <= 0 then
    return "00:00:00";
  else
    hours = string.format("%02.f", math.floor(seconds/3600));
    mins = string.format("%02.f", math.floor(seconds/60 - (hours*60)));
    secs = string.format("%02.f", math.floor(seconds - hours*3600 - mins *60));
    return hours..":"..mins..":"..secs
  end
end

function OnFrame()
  local gui_width = GFGetGuiWidth();
  if (game_state == 0) then
    PopSetFont(3);
    local seconds = math.floor(ready_time/12);
    local str = string.format("Game Starts In: %i seconds", seconds);
    LbDraw_Text((math.floor(ScreenWidth() / 2) - math.floor(string_width(str) / 2)) + math.floor(gui_width / 2), ScreenHeight() - 32, str, 0);
    welcomeBox:Render();
  end
  if (game_state == 1) then
    if (round_time > 0) then
      PopSetFont(3)
      local w = ScreenWidth()
      local ticks_to_seconds = math.floor(round_time/12)
      LbDraw_Text(w-128,16,"Time: " .. SecondsToClock(ticks_to_seconds),0)
    end
    for i,num in ipairs(player_info) do
      local id = player_info[i][2]
      local name = player_info[i][1]
      local w = ScreenWidth()
      local h = ScreenHeight()
      if (w <= 800 and h <= 600) then
        PopSetFont(4)
        LbDraw_Sprite(w-178,(h-30)-(i*12), get_sprite(0,player_info[i][4]))
        LbDraw_Text(w-168,(h-32)-(i*12),"" .. name,0)
        LbDraw_Text(w-88,(h-32)-(i*12),"   : " .. player_info[i][3],0)
      else
        PopSetFont(3)
        LbDraw_ScaledSprite(w-246,(h-30)-(i*16), get_sprite(0,player_info[i][4]),12,12)
        LbDraw_Text(w-228,(h-32)-(i*16),"" .. name,0)
        LbDraw_Text(w-108,(h-32)-(i*16),"   : " .. player_info[i][3],0)
      end
    end
  end
end
