import(Module_Objects)
import(Module_Globals)
import(Module_Game)
import(Module_System)
import(Module_Map)
import(Module_DataTypes)
import(Module_Table)
import(Module_Players)
import(Module_Draw)
import(Module_PopScript)
import(Module_Math)
import(Module_String)
import(Module_Level)

log("Loading script...")

include("TextMsg.lua");

local gs = gsi()
local gns = gnsi()
local person_info = people_type_info()
local spell_info = spells_type_info()

local small_buff = {}
local medium_buff = {}
local large_buff = {}
local lightning_buff = {};

local player_nums = {}
local player_names = {}
local player_spawn_coord = {}
local player_info ={
  {"",0,0,887},
  {"",1,0,914},
  {"",2,0,941},
  {"",3,0,968},
  {"",4,0,1623},
  {"",5,0,1650},
  {"",6,0,1677},
  {"",7,0,1704}
}

local game_state = 0
local round_time = 720*5
local init_turn = gs.Counts.ProcessThings
local message_delay = 48
local ready_time = 12*20;
local mi = 1
local read_me_text = "Welcome to the Luck of the Draw mini-game! In this simple event your main task is to stay at very center of this arena and gain as much points as possible while keeping yourself alive and safe. In this arena there are smoke trails spawned that indicate a lightning strike. Dodge them for your own safety. The closer you are to the center, the more points you gain. Purple being most awarding and green least. Be careful.";
local tut_messages = {
  "Welcome to the Luck Of The Draw mini-game!",
  "Objective of this simple event is to survive as long as possible at the center of arena.",
  "Purple zone awards +144 pts/sec.",
  "Red zone awards +60 pts/sec.",
  "Yellow zone awards +36 pts/sec.",
  "Green zone awards +12 pts/sec.",
  "Be careful! Have fun!"
}

local welcomeBox = TextBox:Create(0, 0, 320, 240);
welcomeBox:FormatString(read_me_text);
welcomeBox:SetHeaderText("Luck of the Draw");

local coord_center = MAP_XZ_2_WORLD_XYZ(50,74)
centre_coord3d_on_block(coord_center)

gns.GameParams.Flags2 = gns.GameParams.Flags2 | (1<<30)

spell_info[2].Cost = 500
person_info[7].LifeIncrease = 255
person_info[7].DefaultLife = 8000
person_info[7].ManaValue = 72

for i=0,7 do
  if (gs.Players[i].DeadCount == 0) then
    local n = get_player_name(i,true)
    player_info[i+1][1] = n
    local c3d = Coord3D.new()
    local shaman = getShaman(i)
    c3d.Xpos = shaman.Pos.D3.Xpos
    c3d.Zpos = shaman.Pos.D3.Zpos
    centre_coord3d_on_block(c3d)
    table.insert(player_nums,i)
    table.insert(player_names,n)
    table.insert(player_spawn_coord,c3d)
    SET_NO_REINC(i,1)
  else
    --Dead or inactive player (bugged?)
    set_player_name(i,"NOPLAYER",true)
    table.insert(player_names,"NOPLAYER")
    player_info[i+1][1] = "NOPLAYER"
    player_info[i+1][3] = -1
  end
end

local DelayedLightning = {};
DelayedLightning.__index = DelayedLightning;

function DelayedLightning:Spawn(_c3d, _count)
  local self = setmetatable({}, DelayedLightning);

  self.Count1 = _count;
  self.Execute = false;
  self.PosC3d = _c3d;
  self.SmokeTrail = ObjectProxy.new();
  self.SmokeTrail:set(0);

  --Create smoke so player can see where lightning will strike gg
  local smoke = createThing(7, 75, 8, _c3d, false, false);
  smoke.u.Effect.Duration = -1;

  self.SmokeTrail:set(smoke.ThingNum);

  return self;
end

function DelayedLightning:IsDone()
  return self.Execute;
end

function DelayedLightning:Process()
  if (not self.Execute) then
    if (self.Count1 > 0) then
      self.Count1 = self.Count1 - 1;
    else
      --createThing(7, 17, 8, self.PosC3d, false, false);
      CREATE_THING_WITH_PARAMS4(7,17,8,self.PosC3d,0,0,0,0)
      DestroyThing(self.SmokeTrail:get());
      self.SmokeTrail:set(0);
      self.Execute = true;
    end
  end
end

-- local function reg_area(_TABLE,c3d,radius,thickness,_alpha)
--   local a = radius
--   local b = thickness
--   local m_idx = MapPosXZ.new()
--   m_idx.Pos = world_coord3d_to_map_idx(c3d)
--   m_idx.XZ.X = m_idx.XZ.X-a
--   m_idx.XZ.Z = m_idx.XZ.Z-a
--   local x = 0
--   local z = 0
--   local num = 1
--   for i=1,a do
--     x = m_idx.XZ.X + (i*2)
--     for j=1,a do
--       z = m_idx.XZ.Z + (j*2)
--       if (i<=b or i>=a-b+1 or j>=a-b+1 or i==a or j<=b or j==a) then
--         local s_idx = MapPosXZ.new()
--         local c_c3d = MAP_XZ_2_WORLD_XYZ(x,z)
--         centre_coord3d_on_block(c_c3d)
--         local me = world_coord3d_to_map_ptr(c_c3d)
--         local spark = createThing(7,61,8,c_c3d,false,false)
--         spark.DrawInfo.Alpha = _alpha
--         spark.Pos.D3.Ypos = spark.Pos.D3.Ypos - 192
--         spark.u.Effect.Duration = -1
--         table.insert(_TABLE,c_c3d)
--       end
--     end
--   end
-- end

for i=0,35 do
  local rotation = (((i * 10)) * math.pi) / 180;
  local c3d = Coord3D.new();
  c3d.Xpos = coord_center.Xpos + math.floor((math.cos(rotation) * (512 * 2)));
  c3d.Zpos = coord_center.Zpos + math.floor((math.sin(rotation) * (512 * 2)));

  local m = createThing(7, 61, 8, c3d, false, false);
  m.Pos.D3.Ypos = 0;
  m.DrawInfo.Alpha = 11;
  m.u.Effect.Duration = -1;
end

for i=0,71 do
  local rotation = (((i * 5)) * math.pi) / 180;
  local c3d = Coord3D.new();
  c3d.Xpos = coord_center.Xpos + math.floor((math.cos(rotation) * (512 * 8)));
  c3d.Zpos = coord_center.Zpos + math.floor((math.sin(rotation) * (512 * 8)));

  local m = createThing(7, 61, 8, c3d, false, false);
  m.Pos.D3.Ypos = 0;
  m.DrawInfo.Alpha = -2;
  m.u.Effect.Duration = -1;
end

for i=0,71 do
  local rotation = (((i * 5)) * math.pi) / 180;
  local c3d = Coord3D.new();
  c3d.Xpos = coord_center.Xpos + math.floor((math.cos(rotation) * (512 * 12)));
  c3d.Zpos = coord_center.Zpos + math.floor((math.sin(rotation) * (512 * 12)));

  local m = createThing(7, 61, 8, c3d, false, false);
  m.Pos.D3.Ypos = 0;
  m.DrawInfo.Alpha = -1;
  m.u.Effect.Duration = -1;
end

for i=0,71 do
  local rotation = (((i * 5)) * math.pi) / 180;
  local c3d = Coord3D.new();
  c3d.Xpos = coord_center.Xpos + math.floor((math.cos(rotation) * (512 * 18)));
  c3d.Zpos = coord_center.Zpos + math.floor((math.sin(rotation) * (512 * 18)));

  local m = createThing(7, 61, 8, c3d, false, false);
  m.Pos.D3.Ypos = point_altitude(c3d.Xpos, c3d.Zpos) - 128;
  m.DrawInfo.Alpha = 1;
  m.u.Effect.Duration = -1;
end

local function toggle_player_spell(_pn,_spell,_arg)
  local ref = gs.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable
  if (_arg) then
    gs.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable = ref | (1 << _spell)
  elseif (not _arg) then
    gs.ThisLevelInfo.PlayerThings[_pn].SpellsAvailable = ref ~ (1 << _spell)
  end
end

local function process_large_areas()
  for i=0,18 do
    local n = G_RANDOM(#large_buff)+1
    --ty brandan no eyerape
    CREATE_THING_WITH_PARAMS4(7,17,8,large_buff[n],0,0,0,0)
  end
end

local function process_medium_areas()
  for i=0,9 do
    local n = G_RANDOM(#medium_buff)+1
    --ty brandan no eyerape
    CREATE_THING_WITH_PARAMS4(7,17,8,medium_buff[n],0,0,0,0)
  end
end

local function spawn_random_lightnings(_coord, _tile_radius)
  local rotation = (G_RANDOM(360) * math.pi) / 180;
  local c3d = Coord3D.new();
  c3d.Xpos = _coord.Xpos + math.floor((math.cos(rotation) * G_RANDOM(512 * _tile_radius)));
  c3d.Zpos = _coord.Zpos + math.floor((math.sin(rotation) * G_RANDOM(512 * _tile_radius)));
  local dl = DelayedLightning:Spawn(c3d, G_RANDOM(20) + 18);
  table.insert(lightning_buff, dl);
end

local function process_delayed_lightnings()
  for i,DL in ipairs(lightning_buff) do
    DL:Process();
  end

  for i,DL in ipairs(lightning_buff) do
    if (DL:IsDone()) then
      table.remove(lightning_buff,i);
    end
  end
end

local function add_score(pn,amt)
  for i in ipairs(player_info) do
    --Make sure playerNum matches with table's one.
    if (player_info[i][2] == pn) then
      player_info[i][3] = player_info[i][3] + amt
      break
    end
  end
end

local function process_shamans()
  for i,pn in ipairs(player_nums) do
    local s = getShaman(pn)
    if (s == nil) then
      createThing(1,7,pn,player_spawn_coord[i],false,false)
    elseif (s ~= nil) then
      if (s.u.Pers.Life > 0) then
        local dist = get_world_dist_xyz(s.Pos.D3, coord_center);
        if (dist < (512 * 2)) then
          add_score(s.Owner, 12);
        elseif (dist < (512 * 8)) then
          add_score(s.Owner, 5);
        elseif (dist < (512 * 12)) then
          add_score(s.Owner, 3);
        elseif (dist < (512 * 18)) then
          add_score(s.Owner, 1);
        elseif (dist > (512 * 18)-1) then
          --nothing...
        end
      end
    end
  end
end

local function decide_winner()
  local ret = -1
  local score = 0
  for i,pnt in ipairs(player_info) do
    if (player_info[i][3] > score) then
      score = player_info[i][3]
      ret = player_info[i][2]
    end
  end
  return ret
end

local function update_leaderboard()
  --Sort by playerNum first then by score num.
  table.sort(player_info, function(a, b) if (a[3] ~= b[3]) then return a[3] < b[3] else return a[2] < b[2] end end)
end

function OnTurn()
  if (gs.Counts.ProcessThings > init_turn + 36) then
    if (game_state == 0) then
      if (ready_time > 0) then
        ready_time = ready_time - 1;
      else
        for i,pn in ipairs(player_nums) do
          createThing(7,49,8,player_spawn_coord[i],false,false)
          toggle_player_spell(pn,2,true)
        end
        game_state = 1
      end
    elseif (game_state == 1) then
      if (round_time > 0) then
        round_time = round_time - 1
        process_shamans()
        process_delayed_lightnings();
        if (gs.Counts.ProcessThings % 4 == 0) then
          update_leaderboard()
          spawn_random_lightnings(coord_center, 18)
        end
        if (gs.Counts.ProcessThings % 36 == 0) then
          spawn_random_lightnings(coord_center, 2)
        end
        if (gs.Counts.ProcessThings % 24 == 0) then
          spawn_random_lightnings(coord_center, 8)
        end
        if (gs.Counts.ProcessThings % 12 == 0) then
          spawn_random_lightnings(coord_center, 12)
        end
      else
        game_state = 2
      end
    elseif (game_state == 2) then
      local winner = decide_winner()
      if (gns.PlayerNum == winner) then
        gns.Flags = gns.Flags | (1<<25)
      else
        gns.Flags = gns.Flags | (1<<26)
      end
      log_msg(8,"Time has ran out! Winner is: " .. player_names[winner+1] .. " - Congratulations!")
      game_state = 3621
    end
  end
end

--[[ took this piece of code from here coz i was lazy to write one myself
https://gist.github.com/jesseadams/791673
]]
local function SecondsToClock(seconds)
  local seconds = tonumber(seconds)

  if seconds <= 0 then
    return "00:00:00";
  else
    hours = string.format("%02.f", math.floor(seconds/3600));
    mins = string.format("%02.f", math.floor(seconds/60 - (hours*60)));
    secs = string.format("%02.f", math.floor(seconds - hours*3600 - mins *60));
    return hours..":"..mins..":"..secs
  end
end

function OnFrame()
  local gui_width = GFGetGuiWidth();
  if (game_state == 0) then
    PopSetFont(3);
    local seconds = math.floor(ready_time/12);
    local str = string.format("Game Starts In: %i seconds", seconds);
    LbDraw_Text((math.floor(ScreenWidth() / 2) - math.floor(string_width(str) / 2)) + math.floor(gui_width / 2), ScreenHeight() - 32, str, 0);
    welcomeBox:Render();
  end
  if (game_state == 1) then
    if (round_time > 0) then
      PopSetFont(3)
      local w = ScreenWidth()
      local ticks_to_seconds = math.floor(round_time/12)
      LbDraw_Text(w-128,16,"Time: " .. SecondsToClock(ticks_to_seconds),0)
    end
    for i,num in ipairs(player_info) do
      local id = player_info[i][2]
      local name = player_info[i][1]
      local w = ScreenWidth()
      local h = ScreenHeight()
      if (w <= 800 and h <= 600) then
        PopSetFont(4)
        LbDraw_Sprite(w-178,(h-30)-(i*12), get_sprite(0,player_info[i][4]))
        LbDraw_Text(w-168,(h-32)-(i*12),"" .. name,0)
        LbDraw_Text(w-88,(h-32)-(i*12),"   : " .. player_info[i][3],0)
      else
        PopSetFont(3)
        LbDraw_ScaledSprite(w-246,(h-30)-(i*16), get_sprite(0,player_info[i][4]),12,12)
        LbDraw_Text(w-228,(h-32)-(i*16),"" .. name,0)
        LbDraw_Text(w-108,(h-32)-(i*16),"   : " .. player_info[i][3],0)
      end
    end
  end
end

log("Script initiated.")
