import(Module_DataTypes)
import(Module_Defines)
import(Module_Draw)
import(Module_Game)
import(Module_Globals)
import(Module_Level)
import(Module_Map)
import(Module_Math)
import(Module_Objects)
import(Module_Players)
import(Module_Person)
import(Module_PopScript)
import(Module_System)
import(Module_String)
import(Module_Table)
import(Module_Sound)

--Includes
include("TextMsg.lua")

--Globals
local gs = gsi()
local gns = gnsi()
local const_spells = spells_type_info()
local const_person = people_type_info()
local launch_turn = gs.Counts.ProcessThings

--Game variables
local game_state = 0
local deathmatch = false
local mi = 1
local prepare_delay = 0
local message_delay = 36
local messages = {
  "Welcome to the Quake Arena mini-game!",
  "Take a little bit of Your time and let me shortly introduce this gamemode.",
  "It's as simple as playing checkers! You have to aim with convert spell!",
  "It'll shoot a deadly laser that will... kill stuff.",
  "And Your objective is to get as many kills as possible within 8 minutes.",
  "Additionally, there is a blast which acts as throwable grenade and a shield powerup!",
  "Get ready to strike!"
}
local read_me_text = "Welcome to the Quake Arena mini-game! Take a little bit of Your time and let me shortly introduce this gamemode. It's as simple as playing checkers! You have to aim with convert spell! It'll shoot a deadly laser that will... kill stuff. And Your objective is to get as many kills as possible within 8 minutes. Additionally, there is a blast which acts as throwable grenade and a shield powerup!"
local timer = 3
local has_given_spells = false
local p_lives = {25,25,25,25,25,25,25,25}
local p_kills = {0,0,0,0,0,0,0,0}
local p_respawn_delay = {60,60,60,60,60,60,60,60}
local round_time = 720*5;
local ready_time = 12*20;

local welcomeBox = TextBox:Create(0, 0, 320, 240);
welcomeBox:FormatString(read_me_text);
welcomeBox:SetHeaderText("Quake Arena");

local map_c3d_respawn_spots = {
  MAP_XZ_2_WORLD_XYZ(108, 34),
  MAP_XZ_2_WORLD_XYZ(52, 12),
  MAP_XZ_2_WORLD_XYZ(36, 224),
  MAP_XZ_2_WORLD_XYZ(50, 172),
  MAP_XZ_2_WORLD_XYZ(98, 158),
  MAP_XZ_2_WORLD_XYZ(136, 194),
  MAP_XZ_2_WORLD_XYZ(148, 248)
}

local powerups_respawn_spots = {
  MAP_XZ_2_WORLD_XYZ(72, 232),
  MAP_XZ_2_WORLD_XYZ(112, 250),
  MAP_XZ_2_WORLD_XYZ(106, 196),
  MAP_XZ_2_WORLD_XYZ(70, 196)
}

for i,c3d in ipairs (powerups_respawn_spots) do
  centre_coord3d_on_block(c3d)
end

--Powerup class
function PowerUp(type,c3d,thingnum)
  local self = {
    respawn_time = 0,
    is_respawning = false
  }

  local _obj_proxy = ObjectProxy.new()
  _obj_proxy:set(thingnum)
  local _type = 0
  local _c3d = c3d
  local _me = world_coord3d_to_map_ptr(c3d)

  function self.ProcessPowerUp()
    if (not _obj_proxy:isNull()) then
      if (gs.Counts.ProcessThings % 2 == 0) then
        local sp = createThing(T_EFFECT,M_EFFECT_SMALL_SPARKLE,8,_c3d,false,false)
        sp.Pos.D3.Ypos = sp.Pos.D3.Ypos + 96
      end
      _me.MapWhoList:processList(function(t)
        if (t.Type == T_PERSON) then
          if (t.Model == M_PERSON_MEDICINE_MAN) then
            if (_type == 0) then
              if (t.u.Pers.u.Owned.ShieldCount == 0) then
                t.u.Pers.u.Owned.ShieldCount = 24
                t.Flags3 = t.Flags3 | TF3_SHIELD_ACTIVE
                self.is_respawning = true
                DestroyThing(_obj_proxy:get())
                return false
              end
            elseif (_type == 1) then
              if (gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_BLAST] < 2) then
                gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_BLAST] = gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_BLAST] + 1
                self.is_respawning = true
                DestroyThing(_obj_proxy:get())
                return false
              end
            end
          end
        end

        return true
      end)
    else
      if (self.respawn_time == 0) then
        if (self.is_respawning) then
          self.respawn_time = 360
          self.is_respawning = false
        else
          local r = G_RANDOM(#powerups_respawn_spots)+1
          _type = G_RANDOM(2)
          local powerup = createThing(T_GENERAL,M_GENERAL_MAPWHO_THING,TRIBE_HOSTBOT,powerups_respawn_spots[r],false,false)
          if (_type == 0) then
            powerup.DrawInfo.DrawNum = 682
            log_msg(TRIBE_HOSTBOT,"Powerup has been spawned!")
          end
          if (_type == 1) then
            powerup.DrawInfo.DrawNum = 1058
            log_msg(TRIBE_HOSTBOT,"Weapon has been spawned!")
          end
          powerup.DrawInfo.Alpha = -16
          powerup.Pos.D3.Ypos = powerup.Pos.D3.Ypos + 192
          _obj_proxy:set(powerup.ThingNum)
          _me = world_coord3d_to_map_ptr(powerup.Pos.D3)
          _c3d = powerup.Pos.D3
        end
      elseif (self.respawn_time > 0) then
        self.respawn_time = self.respawn_time-1
      end
    end
  end

  return self
end

--Kill-list class
function KillList(length)
  local self = {
    messages = {},
    colors = {},
    clear_timer = 0
  }

  local _count = length
  local _clr_nums = {220,244,236,228,146,212,184,136}

  function self.ProcessKillList()
    if (#self.messages > 0) then
      if (self.clear_timer > 0) then
        self.clear_timer = self.clear_timer - 1
        do return end
      end

      self.messages = {}
      self.colors = {}
    end
  end

  function self.ResetTimer()
    self.clear_timer = 60
  end

  function self.AddMsg(string,clr_index)
    table.insert(self.messages,string)
    table.insert(self.colors,_clr_nums[clr_index])
    if (#self.messages > _count) then
      table.remove(self.messages,1)
      table.remove(self.colors,1)
    end
  end

  function self.GetClrBuffer()
    return self.colors
  end

  function self.GetStringBuffer()
    return self.messages
  end

  return self
end

local k_list = KillList(5)

--Fog of war around spawn coords
--[[
for i,c3d in ipairs(map_c3d_respawn_spots) do
  local me = world_coord3d_to_map_ptr(c3d)
  SearchMapCells(CIRCULAR,0,0,4, world_coord3d_to_map_idx(c3d),function(me_s)
    gs.FogOfWar:perm_uncover_all_players(me_s)
    return true
  end)
  set_square_map_params(world_coord3d_to_map_idx(c3d), 5, TRUE);
end
]] --Disabled for now, need some stuff to be done first.

--Graphics
local draw_respawn_time = {false,false,false,false,false,false,false,false}

--Apply edits to constants
const_spells[M_SPELL_CONVERT_WILD].WorldCoordRange = 3072
const_spells[M_SPELL_CONVERT_WILD].Cost = 225
const_spells[M_SPELL_CONVERT_WILD].OneOffMaximum = 1
const_spells[M_SPELL_BLAST].WorldCoordRange = 1024*8
const_spells[M_SPELL_BLAST].Cost = 1225
const_spells[M_SPELL_BLAST].OneOffMaximum = 2
const_person[M_PERSON_MEDICINE_MAN].AttackPower = 5

--Disable reincarnations sites
for i=0,7 do
  SET_NO_REINC(i)
end

--Disable loss
gns.GameParams.Flags2 = gns.GameParams.Flags2 | GPF2_GAME_NO_WIN

--Spells removal function
local function process_spells()
  ProcessGlobalTypeList(T_SPELL, function(t)
      t.u.Spell.TargetThingIdx:set(0)
      t.u.Spell.ShotThingIdx:set(0)
      DestroyThing(t)
    return true
  end)
end

--Find player with most kills
local function find_player_with_most_kills()
  local max_kills = -999999999
  local winner = -1
  for i,value in ipairs(p_kills) do
    if (value > max_kills) then
      max_kills = value
      winner = i-1
    end
  end

  return winner
end

local function find_respawn_spot()
  local r = G_RANDOM(#map_c3d_respawn_spots)+1
  return map_c3d_respawn_spots[r]
end

--Round timer
local function process_round()
  round_time = round_time-1
  if (round_time == 0) then
    deathmatch = true
  end
end

--Respawn function
local function process_respawn()
  for i=0,7 do
    if (gs.Players[i].DeadCount == 0) then
      local s = getShaman(i)
      if (s == nil) then
        if (p_lives[i+1] > 0) then
          if (p_respawn_delay[i+1] > 0) then
            p_respawn_delay[i+1] = p_respawn_delay[i+1] - 1
            draw_respawn_time[i+1] = true
          else
            p_respawn_delay[i+1] = 60
            draw_respawn_time[i+1] = false
            p_lives[i+1] = p_lives[i+1] - 1
            local c3d = find_respawn_spot()
            centre_coord3d_on_block(c3d)
            local sham = createThing(T_PERSON,M_PERSON_MEDICINE_MAN,i,c3d,false,false)
            sham.u.Pers.u.Owned.ShieldCount = 2
            sham.Flags3 = sham.Flags3 | TF3_SHIELD_ACTIVE
            queue_sound_event(sham,SND_EVENT_SHAMAN_RETURN,0)
          end
        end
      end
    end
  end
end

--Disable selecting shamans
ProcessGlobalTypeList(T_PERSON,function(t)
  if (t.Owner >= 0) then
    --log("" .. t.Owner)
    --DestroyThing(t)
    --log("removed")
  end

  return true
end)

local powerups = {}
local first_time_powerup = false

local function process_powerups()
  for i,pow in ipairs(powerups) do
    pow.ProcessPowerUp()
  end
end

function OnTurn()
  if (gs.Counts.ProcessThings > launch_turn + 36) then
    --Tutorial
    k_list.ProcessKillList()
    if (game_state == 0) then
      if (ready_time > 0) then
        ready_time = ready_time - 1;
      else
        game_state = 1
        prepare_delay = gs.Counts.ProcessThings + 24
      end
    elseif (game_state == 1) then
      if (gs.Counts.ProcessThings > prepare_delay) then
        if (timer > 0) then
          prepare_delay = gs.Counts.ProcessThings + 12
          log_msg(TRIBE_HOSTBOT,"" .. timer)
          timer = timer-1
        else
          --Main game
          log_msg(TRIBE_HOSTBOT,"GO!")
          game_state = 2
        end
      end
    elseif (game_state == 2) then
      if (not first_time_powerup) then
        local r = G_RANDOM(#powerups_respawn_spots)+1
        local t = createThing(T_GENERAL,M_GENERAL_MAPWHO_THING,TRIBE_HOSTBOT,powerups_respawn_spots[r],false,false)
        t.DrawInfo.DrawNum = 682
        t.DrawInfo.Alpha = -16
        t.Pos.D3.Ypos = t.Pos.D3.Ypos + 192
        local pow = PowerUp(G_RANDOM(2),t.Pos.D3,t.ThingNum)
        table.insert(powerups,pow)
        first_time_powerup = true
      end
      if (not has_given_spells) then
        for i=0,7 do
          gs.ThisLevelInfo.PlayerThings[i].SpellsAvailable = gs.ThisLevelInfo.PlayerThings[i].SpellsAvailable | (1<<M_SPELL_CONVERT_WILD)
        end
        has_given_spells = true
      end
      process_spells()
      process_respawn()
      process_powerups()
      if (round_time > 0) then
        process_round()
      end

      if (deathmatch) then
        for i=1,8 do
          p_lives[i] = 0
        end
        --Enable back winning/losing
        game_state = 3
      end
    elseif (game_state == 3) then
      local winner = find_player_with_most_kills()
      if (gns.PlayerNum == winner) then
        gns.Flags = gns.Flags | GNS_LEVEL_COMPLETE
      else
        gns.Flags = gns.Flags | GNS_LEVEL_FAILED
      end
      log_msg(TRIBE_HOSTBOT,"Time has ran out! And the winner is: " .. get_player_name(winner,true) .. " Congratulations!")
      game_state = 69
    end
  end
end

function OnCreateThing(t)
  if (t.Type == T_EFFECT) then
    if (t.Model == M_EFFECT_SIMPLE_BLAST) then
      SearchMapCells(SQUARE,0,0,4, world_coord3d_to_map_idx(t.Pos.D3),function(me)
        me.MapWhoList:processList(function(tt)
          if (tt.Type == T_PERSON) then
            if (get_world_dist_xyz(tt.Pos.D3,t.Pos.D3) < 1536) then
              if (tt.Model == M_PERSON_MEDICINE_MAN) then
                local dmg = 65535
                if (tt.u.Pers.u.Owned.ShieldCount > 0) then
                  dmg = 1400
                end
                damage_person(tt,t.Owner,dmg,TRUE)
                createThing(T_EFFECT,M_EFFECT_ORBITER,TRIBE_HOSTBOT,tt.Pos.D3,false,false)
                p_kills[t.Owner+1] = p_kills[t.Owner+1] + 1
                if (tt.u.Pers.Life <= 0) then
                  k_list.ResetTimer()
                  k_list.AddMsg(get_player_name(t.Owner,true) .. " has killed " .. get_player_name(tt.Owner,true),t.Owner+1)
                end
              end
            end
          end
          return true
        end)
        return true
      end)
    end
  end
  if (t.Type == T_SPELL) then
    if (t.Model == M_SPELL_BLAST) then
      gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_BLAST] = gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_BLAST] - 1
      t.u.Spell.TargetThingIdx:set(0)
      t.u.Spell.ShotThingIdx:set(0)
      local s = getShaman(t.Owner)
      if (s ~= nil) then
        local z = (t.Pos.D2.Zpos - s.Pos.D2.Zpos)
        local x = (t.Pos.D2.Xpos - s.Pos.D2.Xpos)
        local ax = math.abs(x)
        local az = math.abs(z)

        if (x > 0 and z > 32767) then
          if (ax > ((256*256)>>1)) then x = -(ax - (256*256)) end
          if (az > ((256*256)>>1)) then z = (az - (256*256)) end
        else
          if (x < 0) then
            if (z >= 0) then
              if (ax > ((256*256)>>1)) then x = -(ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = (az - (256*256)) end
            else
              if (ax > ((256*256)>>1)) then x = -(ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = -(az - (256*256)) end
            end
          else
            if (z >= 0) then
              if (ax > ((256*256)>>1)) then x = (ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = -(az - (256*256)) end
            else
              if (ax > ((256*256)>>1)) then x = (ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = -(az - (256*256)) end
            end
          end
        end
        local angle = math.atan(z, x) * 180 / 3.14
        angle = math.ceil(angle)
        if (angle<0) then
          angle=angle+360
        end

        local rotation = (angle * 3.14) / 180
        local veloc = Coord3D.new()
        local distance_xyz = math.ceil(get_world_dist_xyz(s.Pos.D3,t.Pos.D3)/512)
        --log("Distance: " .. distance_xyz)
        veloc.Xpos = math.ceil((math.cos(rotation) * 32*distance_xyz))
        veloc.Zpos = math.ceil((math.sin(rotation) * 32*distance_xyz))
        --log("X: " .. veloc.Xpos .. " Z: " .. veloc.Zpos)
        local grenade = createThing(T_SHOT,7,s.Owner,s.Pos.D3,false,false)
        grenade.Move.Velocity.X = veloc.Xpos
        grenade.Move.Velocity.Z = veloc.Zpos
        grenade.Move.Velocity.Y = 250
        grenade.Move.SelfPowerSpeed = 128
      end
    end
    if (t.Model == M_SPELL_CONVERT_WILD) then
      gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_CONVERT_WILD] = gs.ThisLevelInfo.PlayerThings[t.Owner].SpellsAvailableOnce[M_SPELL_CONVERT_WILD] - 1
      t.u.Spell.TargetThingIdx:set(0)
      t.u.Spell.ShotThingIdx:set(0)
      local s = getShaman(t.Owner)
      if (s ~= nil) then
        local z = (t.Pos.D2.Zpos - s.Pos.D2.Zpos)
        local x = (t.Pos.D2.Xpos - s.Pos.D2.Xpos)
        local ax = math.abs(x)
        local az = math.abs(z)

        if (x > 0 and z > 32767) then
          if (ax > ((256*256)>>1)) then x = -(ax - (256*256)) end
          if (az > ((256*256)>>1)) then z = (az - (256*256)) end
        else
          if (x < 0) then
            if (z >= 0) then
              if (ax > ((256*256)>>1)) then x = -(ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = (az - (256*256)) end
            else
              if (ax > ((256*256)>>1)) then x = -(ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = -(az - (256*256)) end
            end
          else
            if (z >= 0) then
              if (ax > ((256*256)>>1)) then x = (ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = -(az - (256*256)) end
            else
              if (ax > ((256*256)>>1)) then x = (ax - (256*256)) end
              if (az > ((256*256)>>1)) then z = -(az - (256*256)) end
            end
          end
        end
        local angle = math.atan(z, x) * 180 / 3.14
        angle = math.ceil(angle)
        if (angle<0) then
          angle=angle+360
        end

        local rotation = (angle * 3.14) / 180
        local alt_limit = s.Pos.D3.Ypos+256
        local alt_low_limit = s.Pos.D3.Ypos-(512-128)
        local process = true
        local distance_xyz = get_world_dist_xyz(s.Pos.D3,t.Pos.D3)
        local amount = math.ceil(distance_xyz / 512)
        for i=1,20 do
          if (process) then
            local spawnC3d = Coord3D.new()
            spawnC3d.Xpos = math.ceil(s.Pos.D3.Xpos + (math.cos(rotation) * 512*i))
            spawnC3d.Zpos = math.ceil(s.Pos.D3.Zpos + (math.sin(rotation) * 512*i))
            spawnC3d.Ypos = s.Pos.D3.Ypos
            local res = point_altitude(spawnC3d.Xpos,spawnC3d.Zpos)
            if (res < alt_limit and res > alt_low_limit) then
              alt_limit = res+256
              alt_low_limit = res-(512-128)
              local smok = createThing(T_EFFECT,M_EFFECT_SMALL_SPARKLE,TRIBE_HOSTBOT,spawnC3d,false,false)
              local me = world_coord2d_to_map_ptr(smok.Pos.D2)
              me.MapWhoList:processList(function(t)
                if (t.Type == T_PERSON) then
                  if (t.Owner ~= s.Owner) then
                    if (t.Model > 1 and t.Model < 8) then
                      local dmg = 65535
                      if (t.u.Pers.u.Owned.ShieldCount > 0) then
                        dmg = 500
                      end
                      damage_person(t,s.Owner,dmg,TRUE)
                      createThing(T_EFFECT,M_EFFECT_ORBITER,TRIBE_HOSTBOT,spawnC3d,false,false)
                      p_kills[s.Owner+1] = p_kills[s.Owner+1] + 1
                      if (t.u.Pers.Life <= 0) then
                        k_list.ResetTimer()
                        k_list.AddMsg(get_player_name(s.Owner,true) .. " has killed " .. get_player_name(t.Owner,true),s.Owner+1)
                      end
                    end
                  end
                end

                return true
              end)
              --Whenever shaman walked into laser people thought why it didnt hit, because it was just a visuals.
              --Duration 2 makes it completely instant, which hopefully reveals this illusion.
              smok.u.Effect.Duration = 2
              smok.Pos.D3.Ypos = smok.Pos.D3.Ypos-128
              smok.DrawInfo.Alpha = 0 - (i%2)
            else
              process = false
            end
          end
        end
      end
    end
  end
  if (t.Type == T_SHOT) then
    if (t.Model == M_SHOT_STANDARD) then
      t.u.Shot.TargetThingIdx:set(0)
    end
  end
end

--[[ took this piece of code from here coz i was lazy to write one myself
https://gist.github.com/jesseadams/791673
]]
function SecondsToClock(seconds)
  local seconds = tonumber(seconds)

  if seconds <= 0 then
    return "00:00:00";
  else
    hours = string.format("%02.f", math.floor(seconds/3600));
    mins = string.format("%02.f", math.floor(seconds/60 - (hours*60)));
    secs = string.format("%02.f", math.floor(seconds - hours*3600 - mins *60));
    return hours..":"..mins..":"..secs
  end
end

function OnFrame()
  local gui_width = GFGetGuiWidth();
  if (game_state == 0) then
    PopSetFont(3);
    local seconds = math.floor(ready_time/12);
    local str = string.format("Game Starts In: %i seconds", seconds);
    LbDraw_Text((math.floor(ScreenWidth() / 2) - math.floor(string_width(str) / 2)) + math.floor(gui_width / 2), ScreenHeight() - 32, str, 0);
    welcomeBox:Render();
  end
  if (game_state == 2) then
    local str_buf = k_list.GetStringBuffer()
    if (#str_buf > 0) then
      local clr_buf = k_list.GetClrBuffer()
      for i,string in ipairs(str_buf) do
        local clr = TbColour.new()
        clr.Index = clr_buf[i]
        local offset = math.floor(ScreenWidth()/16)
        SetFont(4)
        LbDraw_Text(math.floor(ScreenWidth()-240),(i*16), "" .. string, clr)
        --LbDraw_PropText(math.floor(ScreenWidth()-180),i*16, "" .. string, COLOUR(11))
      end
    end
    if (round_time > 0) then
      local clr = TbColour.new()
      clr.Index = 155
      local offset = math.floor(ScreenWidth()/16)
      PopSetFont(3)
      local ticks_to_seconds = math.floor(round_time/12)
      LbDraw_Text(math.floor(ScreenWidth()-170),math.floor(ScreenHeight()-50), "Time: " .. SecondsToClock(ticks_to_seconds), clr)
    end

    PopSetFont(3);
    local clr = TbColour.new()
    clr.Index = 155
    local str_1 = string.format("Respawning in: %i", p_respawn_delay[gns.PlayerNum+1])
    local str_2 = string.format("Lives: %i", p_lives[gns.PlayerNum+1])

    if (draw_respawn_time[gns.PlayerNum+1]) then
      LbDraw_Text((math.floor(ScreenWidth() / 2) - math.floor(string_width(str_1) / 2)) + math.floor(gui_width / 2),(math.floor(ScreenHeight()-50)-CharHeight(1)), str_1, clr)
    end
    LbDraw_Text((math.floor(ScreenWidth() / 2) - math.floor(string_width(str_2) / 2)) + math.floor(gui_width / 2),(math.floor(ScreenHeight()-50)), str_2, clr)
  end


end


log("Script initialized!")
--log_msg(TRIBE_HOSTBOT, "Script is loaded!")
